#ifndef __cplusplus
#error Must use C++ for the type apcomplex.
#endif


#if !defined(__APCPLX_H)
#define __APCPLX_H

#if !defined(__APFLOAT_H)
#include "apfloat.h"
#endif


#include <complex.h>


class apcomplex
{
public:
    // Constructors
    apcomplex (apfloat &newre, apfloat &newim = apfloat (new apstruct));
    apcomplex ();

    // Complex manipulations
    friend apfloat real (apcomplex &);          // Real part
    friend apfloat imag (apcomplex &);          // Imaginary part
    friend apcomplex conj (apcomplex &);        // Complex conjugate
    friend apfloat norm (apcomplex &);          // Square of the magnitude
    friend apfloat arg (apcomplex &);           // Angle in the plane

    // Create a complex object given polar coordinates
    friend apcomplex polar (apfloat &mag, apfloat &angle = apfloat (new apstruct));

    // Operator functions
    friend apcomplex operator+ (apcomplex &, apcomplex &);
    friend apcomplex operator+ (apfloat &, apcomplex &);
    friend apcomplex operator+ (apcomplex &, apfloat &);
    friend apcomplex operator- (apcomplex &, apcomplex &);
    friend apcomplex operator- (apfloat &, apcomplex &);
    friend apcomplex operator- (apcomplex &, apfloat &);
    friend apcomplex operator* (apcomplex &, apcomplex &);
    friend apcomplex operator* (apcomplex &, apfloat &);
    friend apcomplex operator* (apfloat &, apcomplex &);
    friend apcomplex operator/ (apcomplex &, apcomplex &);
    friend apcomplex operator/ (apcomplex &, apfloat &);
    friend apcomplex operator/ (apfloat &, apcomplex &);
    friend int operator== (apcomplex &, apcomplex &);
    friend int operator!= (apcomplex &, apcomplex &);
    friend ostream &operator<< (ostream &, apcomplex &);
    apcomplex &operator+= (apcomplex &);
    apcomplex &operator+= (apfloat &);
    apcomplex &operator-= (apcomplex &);
    apcomplex &operator-= (apfloat &);
    apcomplex &operator*= (apcomplex &);
    apcomplex &operator*= (apfloat &);
    apcomplex &operator/= (apcomplex &);
    apcomplex &operator/= (apfloat &);
    apcomplex operator+ ();
    apcomplex operator- ();

    // Other functions
    size_t prec (void);
    void prec (size_t newprec);

    // Implementation
    apfloat re, im;                             // Won't be private
};


// Inline apcomplex functions

inline apcomplex::apcomplex (apfloat &newre, apfloat &newim)
{
    re = newre;
    im = newim;
}

inline apcomplex::apcomplex ()
{
}

inline apcomplex apcomplex::operator+ ()
{
    return *this;
}

inline apcomplex apcomplex::operator- ()
{
    return apcomplex (-re, -im);
}


// Definitions of compound-assignment operator member functions

inline apcomplex &apcomplex::operator+= (apcomplex &z)
{
    re = re + z.re;
    im = im + z.im;

    return *this;
}

inline apcomplex &apcomplex::operator+= (apfloat &d)
{
    re = re + d;

    return *this;
}

inline apcomplex &apcomplex::operator-= (apcomplex &z)
{
    re = re - z.re;
    im = im - z.im;

    return *this;
}

inline apcomplex &apcomplex::operator-= (apfloat &d)
{
    re = re - d;

    return *this;
}

inline apcomplex &apcomplex::operator*= (apcomplex &z)
{
    *this = *this * z;

    return *this;
}

inline apcomplex &apcomplex::operator*= (apfloat &d)
{
    re = re * d;
    im = im * d;

    return *this;
}

inline apcomplex &apcomplex::operator/= (apfloat &d)
{
    *this = *this / d;

    return *this;
}


// Overloaded mathematical functions

apcomplex pow (apcomplex base, unsigned long exp);
apcomplex pow (apcomplex base, unsigned exp);
apcomplex pow (apcomplex base, long exp);
apcomplex pow (apcomplex base, int exp);
apcomplex sqrt (apcomplex z);
apcomplex cbrt (apcomplex z);
apcomplex invroot (apcomplex u, unsigned n);
apfloat abs (apcomplex z);
apcomplex agm (apcomplex a, apcomplex b);
apcomplex exp (apcomplex z);
apcomplex rawlog (apcomplex z);
apcomplex log (apcomplex z);
apcomplex pow (apcomplex z, apfloat w);
apcomplex pow (apfloat z, apcomplex w);
apcomplex pow (apcomplex z, apcomplex w);
apcomplex acos (apcomplex z);
apcomplex acosh (apcomplex z);
apcomplex asin (apcomplex z);
apcomplex asinh (apcomplex z);
apcomplex atan (apcomplex z);
apcomplex atanh (apcomplex z);
apcomplex cos (apcomplex z);
apcomplex cosh (apcomplex z);
apcomplex sin (apcomplex z);
apcomplex sinh (apcomplex z);
apcomplex tan (apcomplex z);
apcomplex tanh (apcomplex z);

// Real trigonometric functions that use complex stuff
apfloat acos (apfloat x);
apfloat asin (apfloat x);
apfloat atan (apfloat x);
apfloat atan2 (apfloat x, apfloat y);
apfloat cos (apfloat x);
apfloat sin (apfloat x);
apfloat tan (apfloat x);


// Definitions of non-member apcomplex functions

inline apfloat real (apcomplex &z)
{
    return z.re;
}

inline apfloat imag (apcomplex &z)
{
    return z.im;
}

inline apcomplex conj (apcomplex &z)
{
    return apcomplex (z.re, -z.im);
}

inline apfloat norm (apcomplex &z)
{
    return z.re * z.re + z.im * z.im;
}

inline apfloat arg (apcomplex &z)
{
    return atan2 (z.re, z.im);
}

inline apcomplex polar (apfloat &mag, apfloat &angle)
{
    return mag * exp (apcomplex (0, angle));
}


// Definitions of non-member binary operator functions

inline apcomplex operator+ (apcomplex &z1, apcomplex &z2)
{
    return apcomplex (z1.re + z2.re, z1.im + z2.im);
}

inline apcomplex operator+ (apfloat &d1, apcomplex &z2)
{
    return apcomplex (d1 + z2.re, z2.im);
}

inline apcomplex operator+ (apcomplex &z1, apfloat &d2)
{
    return apcomplex (z1.re + d2, z1.im);
}

inline apcomplex operator- (apcomplex &z1, apcomplex &z2)
{
    return apcomplex (z1.re - z2.re, z1.im - z2.im);
}

inline apcomplex operator- (apfloat &d1, apcomplex &z2)
{
    return apcomplex (d1 - z2.re, -z2.im);
}

inline apcomplex operator- (apcomplex &z1, apfloat &d2)
{
    return apcomplex (z1.re - d2, z1.im);
}

inline apcomplex operator* (apcomplex &z1, apfloat &d2)
{
    return apcomplex (z1.re * d2, z1.im * d2);
}

inline apcomplex operator* (apfloat &d1, apcomplex &z2)
{
    return apcomplex (d1 * z2.re, d1 * z2.im);
}

inline int operator== (apcomplex &z1, apcomplex &z2)
{
    return (z1.re == z2.re && z1.im == z2.im);
}

inline int operator!= (apcomplex &z1, apcomplex &z2)
{
    return (z1.re != z2.re || z1.im != z2.im);
}
                     

#endif  // __APCPLX_H
